package option_either.example.suche.instance_of;

import option_either.common.exception.UnreachableCodeException;
import option_either.common.exception.ValidationException;


/**
 * Ergebnis-Klasse einer Suche,
 * die je nach
 * Art des Ergebnisses
 * {@link #found},
 * {@link #notFound},
 * {@link #validationError},
 * {@link #error}
 * zurckgibt
 * als Beispiel fr die Nutzung
 * des Option-/Either-Style in Java.
 * 
 * @author Heiner K&uuml;cker
 */
public abstract class InstanceOfSucheResult
{
	/**
	 * Dieser private Konstruktor sorgt dafr,
	 * dass die abstrakte Klasse {@link InstanceOfSucheResult}
	 * nur von inneren Klassen (welche diesen
	 * Konstruktor aufrufen drfen) erweitert
	 * werden kann.
	 */
	private InstanceOfSucheResult()
	{
		super();
	}

	/**
	 * Result-Klasse: Daten gefunden.
	 * <br/><br/>
	 * Diese Klasse ist private,
	 * damit die Client-Seite
	 * gezwungen ist die
	 * {@link Either}-Klasse
	 * zu implementieren, um
	 * an das Ergebnis 
	 * heranzukommen.
	 */
	private static final class Found
	extends InstanceOfSucheResult
	{
		/**
		 * gesuchte Daten.
		 */
		public final String sucheResultStr;

		/**
		 * @param sucheResultStr Daten aus der Suche
		 */
		public Found(
				final String sucheResultStr )
		{
			if ( sucheResultStr == null )
			{
				throw new IllegalArgumentException(
						"sucheResultStr is null" );
			}
			this.sucheResultStr = sucheResultStr;
		}
	}
	
	/**
	 * Result-Klasse: Daten nicht gefunden.
	 * <br/><br/>
	 * Diese Klasse ist private,
	 * damit die Client-Seite
	 * gezwungen ist die
	 * {@link Either}-Klasse
	 * zu implementieren, um
	 * an das Ergebnis 
	 * heranzukommen.
	 */
	private static final class NotFound
	extends InstanceOfSucheResult
	{
		/**
		 * Singleton-Instanz.
		 */
		public static final NotFound instance =
				new NotFound();

		/**
		 * Konstruktor private
		 * for Singleton
		 */
		private NotFound()
		{
			super();
		}
	}

	/**
	 * Result-Klasse: Es ist ein Validierungs-Fehler 
	 * {@link ValidationException}
	 * aufgetreten.
	 * <br/><br/>
	 * Diese Klasse ist private,
	 * damit die Client-Seite
	 * gezwungen ist die
	 * {@link Either}-Klasse
	 * zu implementieren, um
	 * an das Ergebnis 
	 * heranzukommen.
	 */
	private static final class ValidationError
	extends InstanceOfSucheResult
	{
		/**
		 * Fehlermeldung
		 */
		public final String msgStr;

		/**
		 * @param msgStr Fehlermeldung
		 */
		public ValidationError(
				final String msgStr )
		{
			this.msgStr = msgStr;
		}
	}

	/**
	 * Result-Klasse: Es ist ein Fehler aufgetreten.
	 * <br/><br/>
	 * Diese Klasse ist private,
	 * damit die Client-Seite
	 * gezwungen ist die
	 * {@link Either}-Klasse
	 * zu implementieren, um
	 * an das Ergebnis 
	 * heranzukommen.
	 */
	private static final class Error
	extends InstanceOfSucheResult
	{
		/**
		 * Fehlermeldung
		 */
		public final String msgStr;

		/**
		 * @param msgStr Fehlermeldung
		 */
		public Error(
				final String msgStr )
		{
			this.msgStr = msgStr;
		}
	}

	/**
	 * Aufzurufende Methode zum
	 * Verpacken der gefundenen 
	 * Daten.
	 * 
	 * @param sucheResultStr gefundene Daten
	 * @return {@link InstanceOfSucheResult}-Objekt
	 */
	public static InstanceOfSucheResult found(
			final String sucheResultStr )
	{
		return new Found(
				sucheResultStr );
	}

	/**
	 * Aufzurufende Methode zum
	 * Verpacken der Information,
	 * dass keine Daten gefunden
	 * wurde.
	 * 
	 * @return {@link InstanceOfSucheResult}-Objekt
	 */
	public static InstanceOfSucheResult notFound()
	{
		return NotFound.instance;
	}

	/**
	 * Aufzurufende Methode zum
	 * Verpacken der Information,
	 * dass ein Validierungs-Fehler
	 * aufgetreten ist.
	 * 
	 * @param msgStr Fehlermeldungs-String
	 * @return {@link InstanceOfSucheResult}-Objekt
	 */
	public static InstanceOfSucheResult validationError(
			final String msgStr )
	{
		return new ValidationError(
				msgStr );
	}

	/**
	 * Aufzurufende Methode zum
	 * Verpacken der Information,
	 * dass ein unspezifischer Fehler
	 * aufgetreten ist.
	 * 
	 * @param msgStr Fehlermeldungs-String
	 * @return {@link InstanceOfSucheResult}-Objekt
	 */
	public static InstanceOfSucheResult error(
			final String msgStr )
	{
		return new Error(
				msgStr );
	}

	/**
	 * Entweder-Klasse zum
	 * Reagieren je nach Art des
	 * Ergebnis der Suche.
	 * 
	 * @param <R> Type-Parameter, Return-Typ der {@link #execute}-Methode, frei verwendbar in Implementierungs-Klasse
	 */
	abstract public /*non static*/ class Either<R>
	{
		/**
		 * Aufzurufende Methode zum
		 * Reagieren je nach Art des
		 * Ergebnis der Suche.
		 * <br/><br/>
		 * In dieser Methode wird die
		 * Entscheidung von welcher
		 * Art das Ergebnis der Suche
		 * war fr den Client transparent
		 * und fehlersicher vorgenommen.
		 * 
		 * @return Ergebnis der Reaktion
		 */
		public final R execute()
		{
			if ( InstanceOfSucheResult.this instanceof Found )
				// Daten gefunden
			{
				return found(
						( (Found) InstanceOfSucheResult.this ).sucheResultStr );
			}
			else if ( InstanceOfSucheResult.this instanceof NotFound )
			    // keine Daten gefunden
			{
				return notFound();
			}
			else if ( InstanceOfSucheResult.this instanceof ValidationError )
				// Fehler aufgetreten
			{
				return validationError(
						( (ValidationError) InstanceOfSucheResult.this ).msgStr );
			}
			else if ( InstanceOfSucheResult.this instanceof Error )
				// Fehler aufgetreten
			{
				return error(
						( (Error) InstanceOfSucheResult.this ).msgStr );
			}
			else
			{
				// diese Code-Stelle darf nie durchlaufen werden
				throw new UnreachableCodeException();
			}
		}

		/**
		 * Zu implementierende Methode
		 * fr die Reaktion auf
		 * gefundenen Daten.
		 * 
		 * @param sucheResultStr gefundene Daten
		 * @return Ergebnis der Reaktion
		 */
		abstract public R found(
				final String sucheResultStr );

		/**
		 * Zu implementierende Methode
		 * fr die Reaktion auf nicht
		 * gefundenen Daten.
		 * 
		 * @return Ergebnis der Reaktion
		 */
		abstract public R notFound();
		
		/**
		 * Zu implementierende Methode
		 * fr die Reaktion auf einen
		 * aufgetretenen Validierungs-Fehler.
		 * 
		 * @param msgStr Fehlermeldungs-String
		 * @return Ergebnis der Reaktion
		 */
		abstract public R validationError(
				final String msgStr );

		/**
		 * Zu implementierende Methode
		 * fr die Reaktion auf einen
		 * aufgetretenen Fehler.
		 * 
		 * @param msgStr Fehlermeldungs-String
		 * @return Ergebnis der Reaktion
		 */
		abstract public R error(
				final String msgStr );
	}

}
