package option_either.example.suche.visitor;

import option_either.common.exception.ValidationException;


/**
 * Ergebnis-Klasse einer Suche,
 * die je nach
 * Art des Ergebnisses
 * {@link #found},
 * {@link #notFound},
 * {@link #validationError},
 * {@link #error}
 * zurckgibt
 * als Beispiel fr die Nutzung
 * des Option-/Either-Style in Java.
 * 
 * @author Heiner K&uuml;cker
 */
public abstract class VisitorSucheResult
{
	/**
	 * Dieser private Konstruktor sorgt dafr,
	 * dass die abstrakte Klasse {@link VisitorSucheResult}
	 * nur von inneren Klassen (welche diesen
	 * Konstruktor aufrufen drfen) erweitert
	 * werden kann.
	 */
	private VisitorSucheResult()
	{
		super();
	}
	
	/**
	 * Zu implementierende Methode zum Besuchen der konkreten Ausprgung.
	 * 
	 * @param eitherVisitor besuchendes {@link Either}-Objekt
	 * @return return-Wert fr die Methode {@link Either#execute()}
	 */
	abstract protected Object visit(
	        final Either<?> eitherVisitor );

	/**
	 * Result-Klasse: Daten gefunden.
	 * <br/><br/>
	 * Diese Klasse ist private,
	 * damit die Client-Seite
	 * gezwungen ist die
	 * {@link Either}-Klasse
	 * zu implementieren, um
	 * an das Ergebnis 
	 * heranzukommen.
	 */
	private static final class Found
	extends VisitorSucheResult
	{
		/**
		 * gesuchte Daten.
		 */
		public final String sucheResultStr;

		/**
		 * @param sucheResultStr Daten aus der Suche
		 */
		public Found(
				final String sucheResultStr )
		{
			if ( sucheResultStr == null )
			{
				throw new IllegalArgumentException(
						"sucheResultStr is null" );
			}
			this.sucheResultStr = sucheResultStr;
		}

        /**
         * @see VisitorSucheResult#visit
         */
        @Override
        protected Object visit( 
                final Either<?> eitherVisitor )
        {
            // Daten gefunden
            return eitherVisitor.found(
                    this.sucheResultStr );
        }
	}
	
	/**
	 * Result-Klasse: Daten nicht gefunden.
	 * <br/><br/>
	 * Diese Klasse ist private,
	 * damit die Client-Seite
	 * gezwungen ist die
	 * {@link Either}-Klasse
	 * zu implementieren, um
	 * an das Ergebnis 
	 * heranzukommen.
	 */
	private static final class NotFound
	extends VisitorSucheResult
	{
		/**
		 * Singleton-Instanz.
		 */
		public static final NotFound instance =
				new NotFound();

		/**
		 * Konstruktor private
		 * for Singleton
		 */
		private NotFound()
		{
			super();
		}

        /**
         * @see VisitorSucheResult#visit
         */
        @Override
        protected Object visit( 
                final Either<?> eitherVisitor )
        {
            // keine Daten gefunden
            return eitherVisitor.notFound();
        }
	}

	/**
	 * Result-Klasse: Es ist ein Validierungs-Fehler 
	 * {@link ValidationException}
	 * aufgetreten.
	 * <br/><br/>
	 * Diese Klasse ist private,
	 * damit die Client-Seite
	 * gezwungen ist die
	 * {@link Either}-Klasse
	 * zu implementieren, um
	 * an das Ergebnis 
	 * heranzukommen.
	 */
	private static final class ValidationError
	extends VisitorSucheResult
	{
		/**
		 * Fehlermeldung
		 */
		public final String msgStr;

		/**
		 * @param msgStr Fehlermeldung
		 */
		public ValidationError(
				final String msgStr )
		{
			this.msgStr = msgStr;
		}

        /**
         * @see VisitorSucheResult#visit
         */
        @Override
        protected Object visit( 
                final Either<?> eitherVisitor )
        {
            // Validierungs-Fehler aufgetreten
            return eitherVisitor.validationError(
                    this.msgStr );
        }
	}

	/**
	 * Result-Klasse: Es ist ein Fehler aufgetreten.
	 * <br/><br/>
	 * Diese Klasse ist private,
	 * damit die Client-Seite
	 * gezwungen ist die
	 * {@link Either}-Klasse
	 * zu implementieren, um
	 * an das Ergebnis 
	 * heranzukommen.
	 */
	private static final class Error
	extends VisitorSucheResult
	{
		/**
		 * Fehlermeldung
		 */
		public final String msgStr;

		/**
		 * @param msgStr Fehlermeldung
		 */
		public Error(
				final String msgStr )
		{
			this.msgStr = msgStr;
		}

        /**
         * @see VisitorSucheResult#visit
         */
        @Override
        protected Object visit( 
                final Either<?> eitherVisitor )
        {
            // Fehler aufgetreten
            return eitherVisitor.error(
                    this.msgStr );
        }
	}

	/**
	 * Aufzurufende Methode zum
	 * Verpacken der gefundenen 
	 * Daten.
	 * 
	 * @param sucheResultStr gefundene Daten
	 * @return {@link VisitorSucheResult}-Objekt
	 */
	public static VisitorSucheResult found(
			final String sucheResultStr )
	{
		return new Found(
				sucheResultStr );
	}

	/**
	 * Aufzurufende Methode zum
	 * Verpacken der Information,
	 * dass keine Daten gefunden
	 * wurde.
	 * 
	 * @return {@link VisitorSucheResult}-Objekt
	 */
	public static VisitorSucheResult notFound()
	{
		return NotFound.instance;
	}

	/**
	 * Aufzurufende Methode zum
	 * Verpacken der Information,
	 * dass ein Validierungs-Fehler
	 * aufgetreten ist.
	 * 
	 * @param msgStr Fehlermeldungs-String
	 * @return {@link VisitorSucheResult}-Objekt
	 */
	public static VisitorSucheResult validationError(
			final String msgStr )
	{
		return new ValidationError(
				msgStr );
	}

	/**
	 * Aufzurufende Methode zum
	 * Verpacken der Information,
	 * dass ein unspezifischer Fehler
	 * aufgetreten ist.
	 * 
	 * @param msgStr Fehlermeldungs-String
	 * @return {@link VisitorSucheResult}-Objekt
	 */
	public static VisitorSucheResult error(
			final String msgStr )
	{
		return new Error(
				msgStr );
	}

	/**
	 * Entweder-Klasse zum
	 * Reagieren je nach Art des
	 * Ergebnis der Suche.
	 * 
	 * @param <R> Type-Parameter, Return-Typ der {@link #execute}-Methode, frei verwendbar in Implementierungs-Klasse
	 */
	abstract public /*non static*/ class Either<R>
	{
		/**
		 * Aufzurufende Methode zum
		 * Reagieren je nach Art des
		 * Ergebnis der Suche.
		 * <br/><br/>
		 * In dieser Methode wird die
		 * Entscheidung von welcher
		 * Art das Ergebnis der Suche
		 * war fr den Client transparent
		 * und fehlersicher vorgenommen.
		 * 
		 * @return Ergebnis der Reaktion
		 */
		@SuppressWarnings( "unchecked" )
        public final R execute()
		{
		    return (R) visit( this );
		}

		/**
		 * Zu implementierende Methode
		 * fr die Reaktion auf
		 * gefundenen Daten.
		 * 
		 * @param sucheResultStr gefundene Daten
		 * @return Ergebnis der Reaktion
		 */
		abstract public R found(
				final String sucheResultStr );

		/**
		 * Zu implementierende Methode
		 * fr die Reaktion auf nicht
		 * gefundenen Daten.
		 * 
		 * @return Ergebnis der Reaktion
		 */
		abstract public R notFound();
		
		/**
		 * Zu implementierende Methode
		 * fr die Reaktion auf einen
		 * aufgetretenen Validierungs-Fehler.
		 * 
		 * @param msgStr Fehlermeldungs-String
		 * @return Ergebnis der Reaktion
		 */
		abstract public R validationError(
				final String msgStr );

		/**
		 * Zu implementierende Methode
		 * fr die Reaktion auf einen
		 * aufgetretenen Fehler.
		 * 
		 * @param msgStr Fehlermeldungs-String
		 * @return Ergebnis der Reaktion
		 */
		abstract public R error(
				final String msgStr );
	}

}
