package option_either.generator;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;

import option_either.common.exception.UnreachableCodeException;
import option_either.generator.definition.EitherDefinition;
import option_either.generator.definition.EitherVariant;
import option_either.generator.definition.Param;
import option_either.generator.definition.SwitchExecutionMethod;
import option_either.generator.definition.clazz.AbstractTypeParameter;
import option_either.generator.definition.clazz.AbstractTypeVariable;
import option_either.generator.definition.clazz.TypeVariable;
import option_either.generator.definition.clazz.WildcardTypeArgument;
import option_either.util.SeparatedStrBuff;

/**
 * Code-Generator fr Klassen nach dem Either-Muster.
 * 
 * @author Heiner K&uuml;cker
 */
public final class EitherGenerator
{
    /**
     * Vom Benutzer aufzurufende Generierungs-Methode.
     * 
     * @param projectSrcFolderStr Pfad-String zum Verzeichnis mit den Java-Quell-Dateien, Wurzel der Package-Struktur
     * @param destinationPackageStr gewnschtes Ziel-Package als String wie blich mit Punkten abgetrennt
     * @param definition Definition zu generierende Either-Klasse
     * @throws IOException wenn die Ausgabe der generierten Java-Klasse auf Disk/Platte scheitert
     */
    public static void generate(
            final String projectSrcFolderStr ,
            final String destinationPackageStr ,
            final EitherDefinition definition )
    throws IOException
    {
        System.out.println( "generate: " + definition.name );
        
        final String eitherJavaStr =
                generateWithoutFileWrite(
                        destinationPackageStr ,
                        definition ,
                        // generatorClass
                        ( new Exception() ).getStackTrace()[ 1 ].getClassName() );

        writeFile(
                projectSrcFolderStr ,
                destinationPackageStr ,
                definition ,
                eitherJavaStr );
    }

    /**
     * Generierungs-Methode ohne Schreiben der
     * generierten Datei auf Disk, hauptschlich
     * fr Unit-Tests ausgelagert.
     * 
     * @param destinationPackageStr Ziel-Package als String
     * @param definition zu generierende Either-Definition
     * @param generatorClassName Klassen-Name der konkreten Generatorklasse fr Kopf-Kommentar generated by
     * @return String-Inhalt der generierten Java-Klasse
     */
    public static String generateWithoutFileWrite(
            final String destinationPackageStr ,
            final EitherDefinition definition ,
            final String generatorClassName )
    {
        final AbstractTypeParameter[] eitherClassTypeParameterArr;
        {
            final ArrayList<AbstractTypeParameter> eitherClassTypeParameterList;
            if ( definition.typeParameterArr == null )
            {
                eitherClassTypeParameterList = new ArrayList<AbstractTypeParameter>();
            }
            else
            {
                eitherClassTypeParameterList =
                        new ArrayList<AbstractTypeParameter>(
                                Arrays.asList(
                                        definition.typeParameterArr ) );
            }

            eitherClassTypeParameterList.add(
                    new TypeVariable(
                            //description
                            resultTypeparamDescription(
                                    definition ) ,
                            //name
                            "R" ) );

            eitherClassTypeParameterArr =
                    eitherClassTypeParameterList.toArray(
                            new AbstractTypeParameter[ eitherClassTypeParameterList.size() ] );
        }

        final AbstractTypeParameter[] eitherWildcardClassTypeParameterArr;
        {
            final ArrayList<AbstractTypeParameter> eitherWildcardClassTypeParameterList;
            if ( definition.typeParameterArr == null )
            {
                eitherWildcardClassTypeParameterList = new ArrayList<AbstractTypeParameter>();
            }
            else
            {
                eitherWildcardClassTypeParameterList =
                        new ArrayList<AbstractTypeParameter>(
                                Arrays.asList(
                                        definition.typeParameterArr ) );
            }

            eitherWildcardClassTypeParameterList.add(
                    WildcardTypeArgument.instance );

            eitherWildcardClassTypeParameterArr =
                    eitherWildcardClassTypeParameterList.toArray(
                            new AbstractTypeParameter[ eitherWildcardClassTypeParameterList.size() ] );
        }

        final StringBuilder buff = new StringBuilder();

        buff.append( "package " );
        buff.append( destinationPackageStr );
        buff.append( ";\n\n" );

        buff.append( "/**\n" );
        buff.append( " * Generierte Either-Klasse fr\n" );
        buff.append( " * " );
        buff.append( definition.description );
        buff.append( ".\n" );
        
        buff.append(
                generateTypeParamJavadocStr(
                        definition.typeParameterArr ,
                        " * " ) );

        buff.append( " *\n" );
        buff.append( " * Generated by {@link " );
        buff.append( generatorClassName );
        buff.append( "}\n" );
        buff.append( " */\n" );
        buff.append( "abstract public class " );
        buff.append( definition.name );

        buff.append(
                generateTypeParamStr(
                        definition.typeParameterArr ) );

        buff.append( " {\n\n" );
        
        /*
         * abstrakte visit-Methode
         */
        buff.append( "\t/**\n" );
        buff.append( "\t * Zu implementierende visit-Methode nur fr interne Verwendung.\n" );
        buff.append( "\t *\n" );
        buff.append( "\t * @param eitherVisitor Visitor\n" );
        buff.append( "\t * @return frei vom Anwendungsprogrammierer verwendbar\n" );
        buff.append( "\t */\n" );
        buff.append( "\tabstract protected Object visit(Either<?> eitherVisitor);\n\n" );

        for ( final EitherVariant variant : definition.variantArr )
        {
            buff.append(
                    generateVariantClassStr(
                            definition ,
                            variant ,
                            eitherWildcardClassTypeParameterArr ) );
        }
        
        buff.append(
                generateEitherClassStr(
                        definition ,
                        eitherClassTypeParameterArr ) );
        
        /*
         * Factory-Methoden
         */
        for ( final EitherVariant variant : definition.variantArr )
        {
            buff.append(
                    generateFactoryMethodStr(
                            definition ,
                            variant ) );
        }

        buff.append( "}\n" );

        return buff.toString();
    }

    /**
     * Methode zum Generieren der statischen
     * Factory-Methoden je Variante.
     * 
     * @param definition zu generierende Either-Definition
     * @param variant zu generierende Variante
     * @return Java-Code als String
     */
    private static String generateFactoryMethodStr(
            final EitherDefinition definition ,
            final EitherVariant variant )
    {
        final StringBuilder buff = new StringBuilder();

        buff.append( "\t/**\n" );
        buff.append( "\t * ffentliche Factory-Methode.\n" );
        buff.append( "\t *\n" );
        buff.append( "\t * @return " );
        buff.append( variant.description );
        buff.append( "\n" );

        buff.append(
                generateParamDocStr(
                        "\t" ,
                        variant ) );

        buff.append( "\t */\n" );
        buff.append( "\tpublic static " );
        
        buff.append(
                generateTypeParamStr(
                        definition.typeParameterArr ) );
        
        buff.append( " " );
        buff.append( variant.name );
        buff.append(
                generateTypeParamStr(
                        definition.typeParameterArr ) );
        
        buff.append( " " );
        buff.append( firstCharLowercase( variant.name ) );
        buff.append( "(" );

        /*
         * Methoden-Parameter
         */
        buff.append(
                generateParamStr(
                        variant ) );

        buff.append( ") {\n" );

        buff.append( "\t\treturn new " );
        buff.append( variant.name );

        buff.append(
                generateTypeParamStr(
                        definition.typeParameterArr ) );
        
        buff.append( "(" );
        /*
         * Parameter-bergabe an Konstruktor
         */
        boolean isFirstParam = true;
        
        for ( Param param : variant.paramArr )
        {
            if ( isFirstParam )
            {
                isFirstParam = false;
                buff.append( "\n" );
            }
            else
            {
                buff.append( ",\n" );
            }
            buff.append( "\t\t\t\t" );
            buff.append( param.name );
        }
        buff.append( ");\n" );

        buff.append( "\t}\n\n" );

        return buff.toString();
    }

    /**
     * Methode zum Generieren einer inneren
     * Klasse fr die bergebene Variante.
     * 
     * @param definition zu generierende Either-Definition
     * @param variant zu generierende Variante
     * @param eitherWildcardClassTypeParameterArr Type-Parameter-Array der Either-Klasse mit letztem Wildcard-Type-Parameter
     * @return Java-Code als String
     */
    private static String generateVariantClassStr(
            final EitherDefinition definition ,
            final EitherVariant variant ,
            final AbstractTypeParameter[] eitherWildcardClassTypeParameterArr )
    {
        final StringBuilder buff = new StringBuilder();

        buff.append( "\t/**\n" );
        buff.append( "\t * Varianten-Klasse fr " );
        buff.append( variant.description );
        buff.append( ".\n" );
        buff.append(
                generateTypeParamJavadocStr(
                        definition.typeParameterArr ,
                        "\t * " ) );

        buff.append( "\t */\n" );
        buff.append( "\tprivate static final class " );
        buff.append( variant.name );

        if ( variant.typeParameterArr != null &&
                variant.typeParameterArr.length > 0 )
        {
            buff.append( "<" );
            
            final SeparatedStrBuff typeParamStrBuff = new SeparatedStrBuff( ", " );
            
            for ( final AbstractTypeParameter typeParam : variant.typeParameterArr )
            {
                typeParamStrBuff.add( typeParam.toString() );
            }
            
            buff.append( typeParamStrBuff );
            buff.append( ">" );
        }

        buff.append( " extends " );
        buff.append( definition.name );

        buff.append(
                generateTypeParamStr(
                        definition.typeParameterArr ) );

        buff.append( " {\n\n" );

        /*
         * Member
         */
        for ( Param param : variant.paramArr )
        {
            buff.append( "\t\t/**\n" );
            buff.append( "\t\t * " );
            buff.append( param.description );
            buff.append( "\t\n" );
            buff.append( "\t\t */\n" );
            buff.append( "\t\tprivate final " );
            buff.append( param.clazz );
            buff.append( " " );
            buff.append( param.name );
            buff.append( ";\n\n" );
        }
        
        /*
         * Konstruktor
         */
        buff.append( "\t\t/**\n" );
        buff.append( "\t\t * Konstruktor private,\n" );
        buff.append( "\t\t * bitte Factory-Methode \n" );
        buff.append( "\t\t * {@link #" + firstCharLowercase( variant.name ) + "}\n" );
        buff.append( "\t\t * benutzen.\n" );
        buff.append( "\t\t *\n" );

        buff.append(
                generateParamDocStr(
                        "\t\t" ,
                        variant ) );

        buff.append( "\t\t */\n" );
        buff.append( "\t\tprivate " );
        buff.append( variant.name );
        buff.append( "(" );

        /*
         * Konstruktor-Parameter
         */
        buff.append(
                generateParamStr(
                        variant ) );

        buff.append( ") {\n" );

        /*
         * Member-Initialisierung
         */
        for ( Param param : variant.paramArr )
        {
            buff.append( "\t\t\tthis." );
            buff.append( param.name );
            buff.append( "=" );
            buff.append( param.name );
            buff.append( ";\n" );
        }
        buff.append( "\t\t}\n\n" );

        /*
         * visit-Methode
         */
        buff.append( "\t\t/**\n" );
        buff.append( "\t\t * @see " + definition.name + "#visit\n" );
        buff.append( "\t\t */\n" );
        buff.append( "\t\t@Override\n" );
        buff.append( "\t\tprotected Object visit( final Either<?> eitherVisitor ) {\n" );
        buff.append( "\t\t\treturn eitherVisitor." );
        buff.append( firstCharLowercase( variant.name ) );
        buff.append( "(" );

        /*
         * Member-bergabe an Visitor
         */
        boolean isFirstParam = true;
        
        for ( Param param : variant.paramArr )
        {
            if ( isFirstParam )
            {
                isFirstParam = false;
            }
            else
            {
                buff.append( ", " );
            }
            buff.append( "this." );
            buff.append( param.name );
        }
        buff.append( ");\n" );

        buff.append( "\t\t}\n\n" );
        
        /*
         * toString-Methode
         */
        buff.append( "\t\t/**\n" );
        buff.append( "\t\t * @see java.lang.Object#toString()\n" );
        buff.append( "\t\t */\n" );
        buff.append( "\t\t@Override\n" );
        buff.append( "\t\tpublic String toString() {\n" );
        buff.append( "\t\t\treturn this.getClass().getSimpleName()" );

        if ( variant.paramArr.length > 0 )
        {
            buff.append( " +\n" );
            buff.append( "\t\t\t\t\t\"[\" +\n" );
            final SeparatedStrBuff toStrMemberBuff = new SeparatedStrBuff( " + \", \" +\n" );

            for ( Param param : variant.paramArr )
            {
                toStrMemberBuff.add( "\t\t\t\t\t\"" + param.name + "=\" + this." + param.name );
            }

            buff.append( toStrMemberBuff.toString() );
            buff.append( " +\n" );
            buff.append( "\t\t\t\t\t\"]\"" );
        }
        buff.append( ";\n" );
        buff.append( "\t\t}\n\n" );

        // Ende innere Klasse
        buff.append( "\t}\n\n" );

        return buff.toString();
    }

    /**
     * Methode zum Generieren der abstrakten
     * (vom Benutzer/Anwendungsprogrammierer
     *  zu implementierende) Klasse zum
     * Behandeln der jeweils konkreten
     * Variante.
     * 
     * @param definition zu generierende Either-Definition
     * @param eitherClassTypeParameterArr Type-Parameter-Array der Either-Klasse mit letztem R(return)-Type-Parameter
     * @return Java-Code als String
     */
    private static String generateEitherClassStr(
            final EitherDefinition definition ,
            final AbstractTypeParameter[] eitherClassTypeParameterArr )
    {
        final StringBuilder buff = new StringBuilder();

        buff.append( "\t/**\n" );
        buff.append( "\t * Abstrakte Either-Klasse.\n" );
        buff.append( "\t * \n" );
        buff.append( "\t * Diese Klasse muss implementiert werden,\n" );
        buff.append( "\t * um die konkrete Ausprgung des/der\n" );
        buff.append( "\t * " + definition.name + "\n" );
        buff.append( "\t * und die mitgefhrten Werte zu erhalten.\n" );
        buff.append( "\t * \n" );
        buff.append( "\t * @param <R> Type-Parameter, " );

        buff.append(
                resultTypeparamDescription(
                        definition ) );

        buff.append( "\n" );
        buff.append( "\t */\n" );
        // festgelegter Rckgabe-Type-Parameter R
        buff.append( "\tabstract public class Either<R> {\n\n" );

        /*
         * abstrakte Behandlungs-Methoden der einzelnen Varianten
         */
        for ( final EitherVariant variant : definition.variantArr )
        {
            buff.append( "\t\t/**\n" );
            buff.append( "\t\t * Zu implementierende Methode fr \n" );
            buff.append( "\t\t * " + variant.description + ".\n" );
            buff.append( "\t\t * \n" );
            buff.append(
                    generateParamDocStr(
                            "\t\t" ,
                            variant ) );
            buff.append( "\t\t * @return frei vom Anwendungsprogrammierer verwendbar\n" );
            buff.append( "\t\t */\n" );

            buff.append( "\t\tabstract public R " );
            buff.append( firstCharLowercase( variant.name ) );
            buff.append( "(" );

            buff.append(
                    generateParamStr(
                            variant ) );

            buff.append( ");\n\n" );
        }

        if ( definition.switchExecutionMethod.equals(
                SwitchExecutionMethod.EXECUTE_METHOD ) )
        {
            /*
             * execute-Methode
             */
            buff.append(
                    generateExecuteMethod(
                            definition ) );
        }
        else if ( definition.switchExecutionMethod.equals(
                SwitchExecutionMethod.CONSTRUCTOR ) )
        {
            /*
             * Konstruktor
             */
            buff.append(
                    generateResultMemberAndConstructor(
                            definition ) );
        }
        else
        {
            throw new UnreachableCodeException(
                    String.valueOf(
                            definition.switchExecutionMethod ) );
        }

        /*
         * toString-Methode
         */
        buff.append( "\t\t/**\n" );
        buff.append( "\t\t * @see java.lang.Object#toString()\n" );
        buff.append( "\t\t */\n" );
        buff.append( "\t\t@Override\n" );
        buff.append( "\t\tpublic String toString() {\n" );
        buff.append( "\t\t\treturn this.getClass().getSimpleName() +\n" );
        buff.append( "\t\t\t\t\t\"[\" +\n" );
        buff.append( "\t\t\t\t\t\"" + definition.name + ".this=\" + " + definition.name + ".this +" );

        if ( definition.switchExecutionMethod.equals(
                SwitchExecutionMethod.EXECUTE_METHOD ) )
        {
            buff.append( "\n" );
        }
        else
        {
            buff.append(  "\", \" +\n\t\t\t\t\t\"result=\" + this.result +\n" );
        }

        buff.append( "\t\t\t\t\t\"]\";\n" );
        buff.append( "\t\t}\n\n" );

        buff.append( "\t}\n\n" );

        return buff.toString();
    }

    /**
     * 
     * @param definition zu generierende Either-Definition
     * @return Description-String fr den Typ-Parameter Result/Return
     */
    private static String resultTypeparamDescription(
            final EitherDefinition definition )
    {
        return "Type-Parameter, " +
        ( definition.switchExecutionMethod.equals(
                SwitchExecutionMethod.EXECUTE_METHOD )
            ? "Return-Typ der Methode {@link #execute}"
            : "Typ der Member {@link #result}" );
    }

    /**
     * Generieren der execute-Methode.
     * 
     * @param definition zu generierende Either-Definition
     * @return Java-Code als String
     */
    private static String generateExecuteMethod(
            final EitherDefinition definition )
    {
        final StringBuilder buff = new StringBuilder();

        buff.append( "\t\t/**\n" );
        buff.append( "\t\t * Aufzurufende Methode fr konkrete Variante des/der\n" );
        buff.append( "\t\t * " + definition.description + ".\n" );
        buff.append( "\t\t * \n" );
        buff.append( "\t\t * @return frei vom Anwendungsprogrammierer verwendbar\n" );
        buff.append( "\t\t */\n" );
        buff.append( "\t\t@SuppressWarnings( \"unchecked\" )\n" );
        buff.append( "\t\tpublic final R execute() {\n" );
        buff.append( "\t\t\treturn (R) visit( this );\n" );
        buff.append( "\t\t}\n\n" );

        return buff.toString();
    }

    /**
     * Generieren Result-Member und Konstruktor.
     * 
     * @param definition zu generierende Either-Definition
     * @return Java-Code als String
     */
    private static String generateResultMemberAndConstructor(
            final EitherDefinition definition )
    {
        final StringBuilder buff = new StringBuilder();

        buff.append( "\t\t/**\n" );
        buff.append( "\t\t * Result-Member fr Either\n" );
        buff.append( "\t\t */\n" );
        buff.append( "\t\tpublic final R result;\n\n" );

        buff.append( "\t\t/**\n" );
        buff.append( "\t\t * Konstruktor mit Aufruf der Methode fr konkrete Variante des/der\n" );
        buff.append( "\t\t * " + definition.description + ".\n" );
        buff.append( "\t\t */\n" );
        buff.append( "\t\t@SuppressWarnings( \"unchecked\" )\n" );
        buff.append( "\t\tpublic Either() {\n" );
        buff.append( "\t\t\tthis.result = (R) visit( this );\n" );
        buff.append( "\t\t}\n\n" );

        return buff.toString();
    }

    /**
     * Methode zum Generieren des Javadoc
     * fr die Parameter einer Variante.
     * 
     * @param indentStr Einrckungs-String
     * @param variant zu generierende Variante
     * @return Java-Code als String
     */
    private static String generateParamDocStr(
            final String indentStr ,
            final EitherVariant variant )
    {
        final StringBuilder buff = new StringBuilder();

        for ( Param param : variant.paramArr )
        {
            buff.append( indentStr );
            buff.append( " * @param " );
            buff.append( param.name );
            buff.append( " " );
            buff.append( param.description );
            buff.append( "\n" );
        }

        return buff.toString();
    }

    /**
     * Methode zum Generieren der Parameter
     * des Konstruktors bzw der Factory-Methode
     * einer Variante.
     * 
     * @param variant zu generierende Variante
     * @return Java-Code als String
     */
    private static String generateParamStr(
            final EitherVariant variant )
    {
        final StringBuilder buff = new StringBuilder();

        boolean isFirstParam = true;
        
        for ( Param param : variant.paramArr )
        {
            if ( isFirstParam )
            {
                isFirstParam = false;
                buff.append( "\n" );
            }
            else
            {
                buff.append( ",\n" );
            }

            buff.append( "\t\t\t\tfinal " );
            buff.append( param.clazz );
            buff.append( " " );
            buff.append( param.name );
        }

        return buff.toString();
    }

    /**
     * 
     * Hilfsmethode zum Erzeugen der Type-Parameter einer Klasse/Member/Parameter.
     * 
     * @param typeParameterArr Type-Parameter-Array
     * @return generierter Java-Code-String
     */
    private static String generateTypeParamStr(
            final AbstractTypeParameter[] typeParameterArr )
    {
        if ( typeParameterArr == null ||
                typeParameterArr.length == 0 )
        {
            return "";
        }
        else
        {
            final SeparatedStrBuff typeParamStrBuff = new SeparatedStrBuff( ", " );
            
            for ( final AbstractTypeParameter typeParam : typeParameterArr )
            {
                typeParamStrBuff.add( typeParam.toString() );
            }
            
            return "<" + typeParamStrBuff + ">";
        }
    }

    /**
     * Hilfsmethode zum Erzeugen der Javadoc fr die Type-Parameter einer Klasse.
     * 
     * @param typeVariableArr Type-Parameter-Array
     * @param indentStr Einrckung
     * @return generierter Javadoc-String
     */
    private static String generateTypeParamJavadocStr(
            final AbstractTypeVariable[] typeVariableArr ,
            final String indentStr )
    {
        final StringBuilder buff = new StringBuilder();

        if ( typeVariableArr != null &&
                typeVariableArr.length > 0 )
        {
            // Leer-/Trenn-Zeile im Kommentar
            buff.append( indentStr );
            buff.append( "\n" );

            for ( AbstractTypeVariable defiTypeVar : typeVariableArr )
            {
                buff.append( indentStr );
                buff.append(
                        defiTypeVar.toJavadocStr() );
                buff.append( "\n" );
            }
        }

        return buff.toString();
    }

    /**
     * Datei schreiben.
     * 
     * @param projectSrcFolderStr Wurzelverzeichnis des Packages, in welches die generierte Datei gespeichert werden soll
     * @param destinationPackageStr Package-String der generierten Datei, woraus sich das konkrete Verzeichnis fr die generierte Datei ergibt
     * @param definition zu generierende Either-Definition
     * @param eitherJavaFileStr Java-Code-String
     * @throws IOException rethrown
     */
    private static void writeFile(
            final String projectSrcFolderStr ,
            final String destinationPackageStr ,
            final EitherDefinition definition ,
            final String eitherJavaFileStr )
    throws IOException
    {
        /*
         * Ziel-Verzeichnis der Generierung als String
         */
        final String destinationFolderStr =
                projectSrcFolderStr +
                File.separator +
                destinationPackageStr.replace(
                        //oldChar
                        '.' ,
                        //newChar
                        File.separatorChar ) +
                File.separator;

        final File generatedFile =
                new File(
                        destinationFolderStr +
                        File.separator +
                        definition.name +
                        ".java" );

        System.out.println( "write file " + generatedFile );

        final FileWriter fileWriter =
                new FileWriter(
                        generatedFile );

        try
        {
            fileWriter.append( eitherJavaFileStr );
        }
        finally
        {
            fileWriter.close();
        }
    }

    /**
     * Ersten Buchstaben des bergebenen String
     * in Kleinbuchstaben umwandeln.
     * 
     * @param strToFirstCharLowercase
     * @return Ergebnis-String
     */
    private static String firstCharLowercase(
            final String strToFirstCharLowercase )
    {
        if ( strToFirstCharLowercase == null ||
                strToFirstCharLowercase.length() < 1 )
        {
            return strToFirstCharLowercase;
        }
        return
                Character.toLowerCase(
                        strToFirstCharLowercase.charAt( 0 ) ) +
                strToFirstCharLowercase.substring(
                        //beginIndex
                        1 );
    }

}
